#! /usr/bin/python

from twisted.python.components import implements
from twisted.internet import defer
from twisted.python import failure, log

from petmail import util, crypto, interfaces
NotPermitted = interfaces.NotPermitted

class InvalidMessage(Exception):
    """The message was mangled somehow"""

class BaseMessage(object):
    debug = False
    def __init__(self, agent):
        self.agent = agent
        # delivery attributes
        self.when = util.now()  # the nominal creation time of the message
        self.eventlog = [] # log of what processors have done
        self.doDump = False
        self.provides = []
        # recipientEntry, recipientRecord, grantedPermission
        # senderEntry, acceptedRFC822
        self.senderEntry = None

    def startProcessing(self):
        if self.agent.debugProcessing:
            print "startProcessing"
        self.processors = self.getProcessors()
        self.processing = []
        self.done = False
        self.failed = False
        d = self.processingDeferred = defer.Deferred()
        self.process()
        return d
        
    def process(self):
        """Do whatever can be done to the message."""
        # this function will not raise Unsendable exceptions. Instead, it
        # (or a callback) will run self.processingFailed
        debug = self.agent.debugProcessing
        if debug:
            print "process:"
            print " processors %s" % self.processors
            print " provides %s" % self.provides
            print " processing %s" % self.processing

        if not self.processors and not self.processing:
            if debug:
                print " all done"
            return self.processingComplete()
        runnable = self.findRunnableProcesses()
        if not runnable and not self.processing:
            if debug:
                print " no processors are runnable"
            f = failure.Failure
            why = f(interfaces.Unsendable("no processors are runnable"))
            return self.processingFailed(why)

        if debug and not runnable:
            print " weird, nothing to do"
        self.processing.extend(runnable)
        for p in runnable:
            d = defer.maybeDeferred(p.checkMessage, self)
            if self.debug: print " %s gave %s" % (p, d)
            d.addCallback(self.processorComplete, p)
            d.addErrback(self.processorFailed, p)
        if debug:
            print " done"

    def findRunnableProcesses(self):
        # which processors can be run? Removes them from self.processors
        runnable = []
        need = []
        for p in self.processors[:]:
            canRun = True
            for d in p.getDependencies():
                if not d in self.provides:
                    if d not in need: need.append(d)
                    canRun = False
                    break
            needToRun = False
            # empty provides means it must always be run
            processorProvides = p.getProvides()
            if not processorProvides:
                needToRun = True
            else:
                for f in processorProvides:
                    if not f in self.provides:
                        needToRun = True
            if not needToRun:
                if self.agent.debugProcessing:
                    print "don't need to run", p
                self.processors.remove(p)
            elif canRun:
                self.processors.remove(p)
                runnable.append(p)
        if self.agent.debugProcessing:
            print " runnable:", runnable
        return runnable

    def processorComplete(self, res, processor):
        if self.agent.debugProcessing:
            print "processorComplete", processor
        self.processing.remove(processor)
        if not self.failed:
            self.process()
    def processorFailed(self, why, processor):
        if self.agent.debugProcessing:
            print "processorFailed", processor, why
        self.processing.remove(processor)
        self.processingFailed(why)

    def processingComplete(self):
        # this should only be called once, when the last processor succeeds.
        assert not self.done
        self.done = True
        self.processingDeferred.callback(self)
    def processingFailed(self, why):
        # this may be called multiple times, once per failing processor
        log.msg("processing step failed:", why)
        if self.failed:
            if self.agent.debugProcessing:
                print "processingFailed >1 times"
            return
        self.done = self.failed = True
        self.processingDeferred.errback(why)

    def provide(self, which):
        if which not in self.provides:
            self.provides.append(which)

    def hasNode(self, name):
        return self.headers.has_key(name)
    def getNode(self, name, default=None):
        return self.headers.get(name, default)

    def log(self, message):
        self.eventlog.append(message)
        if self.debug:
            print message

class Message(BaseMessage):
    """This object represents a single message being sent to some individual
    recipient. Its life-cycle is as follows:

     1: Created when the user hits the 'Send' button on their MUA,
        containing an rfc822-formatted header+body pair.
     2: The recipient's AddressBookEntry is looked up.
     3: .encode() is called with the AddressBookEntry. This returns a string
        which should be given to the transport layer for delivery to the
        recipient.
     4: The Message is stored in a list of outstanding messages (indexed by
        messageID). An ack from the far end will remove it from the list. If
        it stays on the list for too long (5 days), the user is notified
        that the message did not make it through.


    Useful attributes of this object:

    .senderEntry: the OwnerEntry which is sending the message
    .recipientAddress: set iff the user is writing to a new recipient for
                       which they only have an address, else None.

    .recipientEntry: at the start of processing, this is only set if the
                     user is writing to a correspondent in their
                     addressbook. After the RecipientEntry processor is run,
                     it will be set (and the entry added to the user's
                     AddressBook) if an rfc822 message is being sent to
                     them.

    .recipientRecord: set when AddressLookup turns .recipientAddress or
                      .recipientEntry into an idRecord
    .transport: an ITransport object capable of sending this message

    .provides is a list of things that have been done to this message:

    senderEntry: .senderEntry has been populated # TODO: require this
    recipientRecord: .recipientRecord has been populated
    recipientEntry: .recipientEntry has been populated
    grantedPermission: return permission has been granted

    """
    __implements__ = interfaces.IMessage

    def __init__(self, agent):
        BaseMessage.__init__(self, agent)
        self.messageID = util.generateMessageID()
        self.recipientAddress = None
        self.recipientRecord = None # 'recipientRecord'
        self.recipientEntry = None # 'recipientEntry'
        self.transport = None # the transport used to send the message
        self.worthSending = False
        self.sent = None # the time this message was sent. TODO: set it
        self.headers = {} # indexed by name

    def setMessageRecipient(self, name, owner):
        # this can accept a pet name 
        book = self.agent.getAddressBook(owner.ID)
        r = book.lookupPetName(name)
        # and addresses
        if not r:
            r = book.lookupAddress(name)
        if r:
            self.setRecipientEntry(r)
        else:
            self.recipientAddress = name # will be looked up later

    def setRecipientRecord(self, r):
        assert implements(r, interfaces.IIDRecord)
        self.recipientRecord = r
        self.provide('recipientRecord')

    def setRecipientEntry(self, e):
        assert implements(e, interfaces.IAddressBookEntry)
        self.recipientEntry = e
        self.provide('recipientEntry')
        self.setRecipientRecord(e.idRecord)

    def setSenderEntry(self, e):
        assert implements(e, interfaces.IOwnerEntry)
        self.senderEntry = e
        self.provide('senderEntry')

    def getProcessors(self):
        return self.agent.getOutboundProcessors()

    def addNode(self, name, value):
        assert (name not in self.headers)
        self.headers[name] = value

    def addRFC822(self, body): #, us, recipient, body, addrbook=None):
        """Add an RFC822 message section to the Message."""

        # construct the header
        header = {}
        #header['bodyHash'] = util.hash(body)
        #header['bodyLen'] = len(body)
        ##header['senderID'] = us.ID
        ##header['messageID'] = util.generateMessageID()
        ##header['timestamp'] = self.when
        header['tickets'] = []
        header['body'] = body

        self.addNode("rfc822", header)
        self.worthSending = True

    def send(self):
        return self.agent.sendMessage(self)

    def addTicket(self, ticket):
        # ticket is a signed blob
        assert(self.headers['rfc822'])
        self.headers['rfc822']['tickets'].append(ticket)

    def encodeHeaders(self):
        return util.encode(self.headers)

    def deliver(self):
        d = self.encrypt(self.senderEntry.privkey,
                           self.recipientRecord.pubkey)
        d.addCallback(self.transport.send)

    def encrypt(self, us, recipient):
        """Returns a Deferred with the encrypted form of this message.
        """
        if not isinstance(us, crypto.PrivateKey):
            print "us", us
        assert isinstance(us, crypto.PrivateKey)
        assert isinstance(recipient, crypto.PublicKey)

        encoded = util.encode(self.headers)
        return recipient.encryptAndSign(us, encoded)

    def dump(self):
        if not self.doDump:
            return
        print "message dump:"
        print " worthSending:", self.worthSending
        keys = self.headers.keys()
        keys.sort()
        print "", ",".join(keys)
        for m in self.eventlog:
            print " %s" % m


class ReceivedMessage(BaseMessage):
    """A ReceivedMessage object is created for each incoming message.

    The .process method is run on it to check permissions. Once this
    succeeds, the following attributes will be available on the
    ReceivedMessage object:

    TODO: add the permission with which the message was accepted

    Useful attributes of this object which are set early:
    .agent: the agent which is receiving the message
    .encrypted: a file object (or StringIO) with the original encrypted form
    .recipientEntry: the OwnerEntry that is the target of this message
    .headers: the unserialized contents of the message

    Attributes which are set by various processing steps (and the .provides
    tags which indicate they are available):

    .allowRFC822 [checkedRFC822]: a boolean indicating that we are willing
    to accept an rfc822 message from this ReceivedMessage

    .allowNewSenderCreation [checkedNewSenderCreation]: a boolean indicating
    that we should be willing to create a new AddressBookEntry for the
    sender of this message. TODO: other Processors should be able to provide
    this flag, so we may need some kind of dependency-AND logic here.

    .senderEntry: always present, but it is only provisional until
    [senderEntry] is provided. Before that point it will be None unless the
    message has a valid signature from a previously-known sender. It may be
    set later (after [senderEntry]) only when a new sender has been accepted
    into the address book.
    
    """

    __implements__ = interfaces.IReceivedMessage

    def __init__(self, agent, encrypted):
        """If any part of the message is invalid, this will raise
        InvalidMessage or UndecryptableError. If the sender does not have
        permission, then it will raise NotPermitted."""

        BaseMessage.__init__(self, agent)
        self.encrypted = encrypted

    def decrypt(self):
        """I decrypt my crypttext and set the following attributes:

        .recipientEntry: the OwnerEntry of the message recipient (to whom
        the message was encrypted)

        .headers: the plaintext message dictionary

        .senderEntry: this will point to an AddressBookEntry if the message
        has a valid signature from a known sender (someone who is already in
        our AddressBook). It will be None if the message had no signature or
        was signed by a sender we don't already know (such as a new sender).

        I return a Deferred which will fire with the .recipientEntry .
        """

        ca = self.agent.crypto
        d = ca.decrypt(self.encrypted)
        d.addCallback(self._decrypt2)
        return d

    def _decrypt2(self, results):
        recipientKey, senderID, plaintext = results
        try:
            recipientID = recipientKey.getID()
            self.recipientEntry = self.agent.getSender(recipientID)
        except KeyError:
            print "received message was meant for someone else"
            raise

        self.senderEntry = None
        if senderID:
            book = self.agent.getAddressBook(recipientID)
            self.senderEntry = book.lookupID(senderID)

        self.headers = util.decode(plaintext)

        return self.recipientEntry

    def getProcessors(self):
        return self.agent.getInboundProcessors()

    def dump(self):
        if not self.doDump:
            return
        print "Message processed"
        if self.senderEntry:
            print " From:", self.senderEntry.getName()
        if "checkedRFC822" in self.provides and self.allowRFC822 and \
            self.getNode("rfc822"):
            l = len(self.getNode('rfc822')['body'])
            print " RFC822 len(%d) accepted" % l
        for m in self.eventlog:
            print " %s" % m
        
