#! /usr/bin/python

from twisted.application import service
from twisted.python.components import implements
from twisted.python import log

from petmail import crypto, transport, util, interfaces, permission

class OwnerEntry(service.MultiService):
    """I hold information about one of the owner's identities. The agent
    holds a collection of these objects. Encoded copies will be published to
    the Name Servers.

    Every time you finish changing something about this record, you must
    call .publish() to push the changes out.

    All OwnerEntry objects are children of the Agent, so they can get at it
    with self.parent
    """

    __implements__ = (interfaces.IOwnerEntry,
                      service.MultiService.__implements__)
    grantReturnPermission = True # tested by processors.GrantReturnPermission
    grantInboundPermission = True #  by processors.GrantInboundPermission

    def __init__(self):
        service.MultiService.__init__(self)
        # .name, .running, and .parent are reserved for use by MultiService

        self.myName = None # local pet name for this identity
        self.publishers = []
        self.inlets = []
        self.lastModified = util.now()
        self.privkey = None

        # fields which get published
        self.ID = None # derived from the pubkey
        self.names = "" # human names, affiliations: multiline text string
        self.addresses = [] # addresses which should resolve to this record
        # fields which are published in some indirect form
        self.genericPermission = None
        self.expirationTime = 120 # one month is more reasonable

    def __repr__(self):
        return "<%s id=%s>" % (self.__class__.__name__, self.ID)

    def modified(self):
        self.lastModified = util.now()

    def setKey(self, privkey):
        assert isinstance(privkey, crypto.PrivateKey)
        self.privkey = privkey
        self.ID = privkey.getID()
        self.modified()

    def changeKey(self):
        raise NotImplemented
        self.modified()

    def setLocalName(self, name):
        self.myName = name
        # not published

    def setName(self, name):
        self.names = name
        self.modified()

    def getName(self):
        if self.myName:
            return self.myName
        if self.names and self.names != "":
            return self.names
        return self.ID

    def addAddress(self, addr):
        self.addresses.append(addr)
        self.modified()

    def publish(self):
        for p in self.publishers:
            for a in self.addresses:
                p.publish(a, self)

    def setGenericPermission(self, perm):
        """Set .genericPermission to None to indicate no permission at all,
        or to an OwnerPermission to set up TicketServers and such."""
        if perm:
            assert implements(perm, interfaces.IOwnerPermission)
        self.genericPermission = perm
        self.modified()

    def grantPermissionTo(self, sender, message):
        assert implements(sender, interfaces.IAddressBookEntry)
        # TODO: This gives full permission. The default ought to be
        # configurable.
        # TODO: record 'message' as the reason for the permission
        if sender.theirPermission:
            log.msg("OwnerEntry.grantPermissionTo: they already it")
            return
        p = permission.OwnerPermission()
        sender.setTheirPermission(p)

    def setExpirationTime(self, t):
        self.expirationTime = t
        self.modified()

    def receiveMessage(self, encrypted):
        self.parent.receiveMessage(encrypted)

    def addInlet(self, inlet):
        assert implements(inlet, interfaces.IInlet)
        self.inlets.append(inlet)
        inlet.setServiceParent(self)
        self.modified()

    def removeInlet(self, inlet):
        self.inlets.remove(inlet)
        inlet.disownServiceParent()
        self.modified()

    def setInlet(self, inlet):
        """This presumes exactly one inlet. 'inlet' may be None."""
        assert len(self.inlets) <= 1
        if inlet in self.inlets:
            return
        if self.inlets:
            self.removeInlet(self.inlets[0])
        if inlet:
            self.addInlet(inlet)
        self.modified()

    def encode(self):
        """I return a Deferred which will fire with the encoded form of this
        record.
        """
        d = self.privkey.getKey()
        d.addCallback(self._encode)
        return d

    def _encode(self, pubkey):
        d = {'ID': self.ID,
             'pubkey': pubkey,
             'name': self.names,
             'addresses': self.addresses,
             'permission': None,
             'expires': util.now() + self.expirationTime,
             'transports': [inlet.encode() for inlet in self.inlets],
             }
        if self.genericPermission:
            d['permission'] = self.genericPermission.encode()
        return util.encode(d)

    def sign(self):
        """Return a Deferred which fires with a signed string ready for
        uploading to a server or including in a a message. This is the
        serialized form of the IDRecord, and can be passed to
        addressbook.IDRecord to create such an object."""

        d = self.encode()
        d.addCallback(self._sign)
        return d

    def _sign(self, encoded):
        return self.privkey.sign(encoded)

