#! /usr/bin/python

"""
Define several built-in Processors to handle messages.
"""

"""
Message Processing (outbound):

Messages are created and populated with nodes, then handed to the Agent. The
Agent gives them a list of Processors that get to run (stored in
.processors).

Each processor has .dependencies and .provides lists to provide a sequence
in which they want to be run. Any given processor should be skipped if
everything it .provides is already present in the message. A processor
should not be run until its .dependencies are all provided by the message.

Each Processor can do one of three things (this might be merged into just
one thing later):

 return True, indicating that the processor has done its job immediately
 return a Deferred, which will callback or errback when the job is done
 raise Unsendable

If the Processor returns a Deferred, it is added to the message's
.processing list. When the Deferred completes, it will be removed from that
list.

Message.process figures out all Processors which can be run and starts them.
It

When the message begins processing, the agent puts in on a .processing list. 
"""


from twisted.internet import defer

from petmail import interfaces, util, addressbook
from petmail.interfaces import Unsendable, NotPermitted

class BaseProcessor(object):
    __implements__ = interfaces.IProcessor
    dependencies = []
    provides = []
    def getDependencies(self):
        return self.dependencies
    def getProvides(self):
        return self.provides
    def configureSender(self, sender):
        raise Unimplemented
    def configureRecipient(self, recipient):
        raise Unimplemented

class OutboundProcessor(BaseProcessor):
    __implements__ = interfaces.IOutboundProcessor
    def checkMessage(self, message):
        if message.agent.debugProcessing:
            print "doing", self.__class__.__name__
        for d in self.dependencies:
            if not d in message.provides:
                raise AssertionError("'%s' not in .provides=%s" \
                                     % (d, message.provides))

class InboundProcessor(BaseProcessor):
    __implements__ = interfaces.IInboundProcessor
    def checkMessage(self, message):
        if message.agent.debugProcessing:
            print "doing", self.__class__.__name__
        for d in self.dependencies:
            assert d in message.provides, "%s not provided" % d



class Delivery(OutboundProcessor):
    def checkMessage(self, message):
        """Make sure the headers have a MessageID node."""

        OutboundProcessor.checkMessage(self, message)
        if not message.hasNode("messageID"):
            message.addNode("messageID", message.messageID)
        return True

class AddressLookup(OutboundProcessor):
    provides = ["recipientRecord"]
    def checkMessage(self, message):
        """If the message has a recipientAddress but not a recipientRecord,
        do an address lookup."""
        OutboundProcessor.checkMessage(self, message)
        if 'recipientRecord' in message.provides:
            return True
        if message.recipientEntry:
            message.setRecipientRecord(message.recipientEntry.idRecord)
            return True
        if message.recipientAddress:
            d = message.agent.retrieveAddress(message.recipientAddress)
            d.addCallback(self.gotRecord, message)
            d.addErrback(self.lookupFailed, message)
            return d
        raise Unsendable, "no address to look up"
    def gotRecord(self, record, message):
        if not record:
            raise Unsendable("unable to look up address")
        message.recipientRecord = record
        message.provide('recipientRecord')
        message.log("AddressLookup returned %s" % record)
        return True
    def lookupFailed(self, why, message):
        message.provide('recipientRecord') # lookup won't succeed
        return why

class RecipientEntry(OutboundProcessor):
    dependencies = ["recipientRecord", "senderEntry"]
    provides = ["recipientEntry"]
    def checkMessage(self, message):
        """If we are sending an rfc822 message to this recipient, create an
        address book entry for them (if necessary).
        """
        OutboundProcessor.checkMessage(self, message)
        if message.recipientEntry:
            # already there
            return True
        message.provide("recipientEntry")
        if not message.getNode('rfc822'):
            # automated message
            return True
        if not message.senderEntry:
            # could just return, but I want to know when this might happen
            raise Unimplemented("when can this happen?")
        
        message.log("installing outbound AddressBook Entry")
        book = message.agent.getAddressBook(message.senderEntry.ID)
        e = addressbook.AddressBookEntry()
        e.idRecord = message.recipientRecord
        message.recipientEntry = e
        book.addEntry(e, message)
        return True

class Transport(OutboundProcessor):
    dependencies = ["recipientRecord"]
    def checkMessage(self, message):
        """If the message does not have a Transport, try to create one."""
        OutboundProcessor.checkMessage(self, message)
        if message.transport:
            return True
        if not message.recipientRecord:
            raise Unsendable, "cannot get a Transport without an IDRecord"
        #import pdb; pdb.Pdb().set_trace()
        if message.agent.debugProcessing: print " getting a transport"
        d = message.recipientRecord.getTransport(message.agent, message)
        if d:
            d.addCallback(self.gotTransport, message)
            d.addErrback(self.failed, message)
        return d
    def gotTransport(self, transport, message):
        message.log("added transport %s" % transport)
        message.transport = transport
        return True
    def failed(self, why, message):
        return why


class UpdateTheirIDRecord(OutboundProcessor):
    dependencies = ['senderEntry', 'recipientEntry']
    def checkMessage(self, message):
        """If their copy of my ID Record is out of date, include a copy."""
        OutboundProcessor.checkMessage(self, message)
        if not message.recipientEntry:
            return True  # one-off recipient? reply?
        lastSent = message.recipientEntry.lastSentOwnerEntry
        lastModified = message.senderEntry.lastModified
        message.log("lastSent %s, lastModified %s" % (lastSent, lastModified))
        if lastSent < lastModified:
            d = message.senderEntry.sign()
            d.addCallback(self.gotSigned, message)
            return d
        return True

    def gotSigned(self, signed, message):
        message.addNode("senderIDRecord", signed)
        message.worthSending = True
        message.recipientEntry.lastSentOwnerEntry = util.now()
        # TODO: update lastSent when ack is received
        message.log("included copy of our IDRecord")
        return True

class GrantReturnPermission(OutboundProcessor):
    dependencies = ['recipientEntry']
    provides = ['grantedPermission']
    def checkMessage(self, message):
        """When sending an rfc822 message, it is polite to give the
        recipient permission to respond."""

        OutboundProcessor.checkMessage(self, message)

        # by the time this processor is done, they'll have whatever
        # permission they're going to get
        message.provide("grantedPermission")

        if not message.hasNode("rfc822"):
            return True
        owner = message.senderEntry
        if not owner:
            return
        e = message.recipientEntry
        if not e:
            return True # we can't give them permission if we aren't going
                        # to bother remembering who they are
        if e.theirPermission:
            return True # they already have permission
        if e.doNotGrantReturnPermission:
            message.log("not granting return permission: we don't like them")
            return True
        if not owner.grantReturnPermission:
            message.log("not granting return permission: we don't do that")
            return True
        message.log("granting return permission")
        owner.grantPermissionTo(e, message)
        return True

class UpdateTheirPermission(OutboundProcessor):
    dependencies = ["recipientEntry", "grantedPermission"]
    def checkMessage(self, message):
        """If the recipient has Permission (the 'theirPermission' field of
        their address book entry), and the timestamps indicate that their
        copy is out of date, I will insert a PermissionSlip into the
        outbound message."""
        OutboundProcessor.checkMessage(self, message)
        if not message.recipientEntry:
            if message.agent.debugProcessing:
                print "no recipientEntry, so no permission to update"
            return True # no permission to be granted
        e = message.recipientEntry
        if not e.theirPermission:
            if message.agent.debugProcessing:
                print "no specific permission, so nothing to update"
            return True # no permission to be granted
        p = e.theirPermission
        if message.agent.debugProcessing:
            print "last sent %d, lastModified %d" % (e.lastSentTheirPermission,
                                                     p.lastModified)
        if e.lastSentTheirPermission < p.lastModified:
            if message.agent.debugProcessing:
                print "updating them"
            message.log("updating their permission")
            message.addNode("yourPermission", e.encodePermissionSlip())
            message.worthSending = True
            e.lastSentTheirPermission = util.now()
            # TODO: update p.lastSent when ack is received. Retransmits.
            if message.agent.debugProcessing:
                print "did update"
        return True

class RFC822Permission(OutboundProcessor):
    dependencies = ['recipientRecord', 'recipientEntry']
    def checkMessage(self, message):
        """Outbound rfc822 messages require Permission."""
        OutboundProcessor.checkMessage(self, message)
        node = message.getNode("rfc822")
        if not node:
            return True
        if not message.recipientRecord:
            raise Unsendable, "need ID Record to determine Permission"
        e = message.recipientEntry
        # do we need any kind of permission for this message?
        p = None
        if e and e.ourPermission:
            p = e.ourPermission
            ptype = "(specific)"
            if message.agent.debugProcessing:
                print "Using Specific Permission, IMG", p.minIMG
        else:
            p = message.recipientRecord.genericPermission
            ptype = "(generic)"
            if message.agent.debugProcessing:
                print "Using Generic Permission", p
        if not p:
            message.log("rfc822 rejected: no permission")
            raise Unsendable, "this recipient does not accept new senders"
        reqs = p.checkOutboundMessage(message)
        assert type(reqs) == type([])
        # checkMsg will raise a message.NotPermitted exception if a size
        # limit is violated. It will return requirements if there are
        # tickets we must obtain before sending
        if not reqs:
            p.sendOutboundMessage(message)
            message.log(ptype + " rfc822 allowed through: no tickets needed")
            return True
        # we need to fulfill this requirement
        if message.agent.debugProcessing:
            print "have reqs", reqs
        dl = []
        tas = message.agent.getTicketAgents()
        if not tas:
            raise Unsendable, ptype + \
                  " message needs tickets but we have no agents to get them"
        for ta in tas:
            deferreds = ta.handleRequirements(reqs, message)
            for d in deferreds:
                if isinstance(d, defer.Deferred):
                    d.addCallback(self.gotTicket, message, p)
                    d.addErrback(self.failed, message)
                    dl.append(d)
        # TODO: make sure the requirements were actually handled. This
        # probably involves having each ticketAgent tell us which
        # requirements they took care of (or started taking care of)
        if dl:
            # when all these Deferreds fires, the message is ready to go
            dlist = defer.DeferredList(dl, fireOnOneErrback=True)
            # the DeferredList is taking ownership of errors in the
            # individual Deferreds. To make sure no errors get left over,
            # add an errback which will drop them.
            [d.addErrback(lambda why: None) for d in dl]
            return dlist
        return True

    def gotTicket(self, ticket, message, p):
        if message.agent.debugProcessing:
            print "gotTicket", ticket
        node = message.getNode('rfc822')
        if not node.has_key("tickets"):
            node["tickets"] = []
        node["tickets"].append(ticket) # TODO: maybe serialize here?
        message.log("adding ticket")
        p.sendOutboundMessage(message)
        return True
    def failed(self, why, message):
        return why


allOutbound = [Delivery, AddressLookup, RecipientEntry,
               Transport, UpdateTheirIDRecord,
               GrantReturnPermission, 
               UpdateTheirPermission, RFC822Permission,
               ]

############################################################

# .senderEntry must be looked up first, but set to None if they are unknown

class CheckRFC822Permission(InboundProcessor):
    provides = ["checkedRFC822", "checkedNewSenderCreation"]

    def done(self, message):
        message.provide("checkedRFC822")
        message.provide("checkedNewSenderCreation")
        return

    def checkMessage(self, message):
        """Incoming rfc822 messages require valid permission."""
        InboundProcessor.checkMessage(self, message)
        a = message.agent
        debug = a.debugProcessing

        message.allowRFC822 = False
        message.allowNewSenderCreation = False
        # TODO: remove that =False if AND-dependencies need it

        if not message.getNode("rfc822"):
            return self.done(message)
        # they use genericPermission if:
        #  A: we don't know who they are
        #  B: we know them but haven't given them specific permission
        if message.senderEntry and message.senderEntry.theirPermission:
            p = message.senderEntry.theirPermission
            ptype = "(specific)"
        else:
            p = message.recipientEntry.genericPermission
            ptype = "(generic)"
        if not p:
            if debug:
                print " rejected, no permission available"
            message.log("rejected rfc822 message: no permission")
            return self.reject(message, NotPermitted("no permission"))

        if debug:
            print " checking", p
        d = defer.maybeDeferred(p.checkInboundMessage, message)
        d.addCallback(self.didCheck, ptype, p, message)
        d.addErrback(self.checkFailed, message)
        return d

    def didCheck(self, why, ptype, p, message):
        debug = message.agent.debugProcessing
        why = "%s %s" % (ptype, why)
        # if we get here, the message is accepted
        if debug:
            print "  accepted"

        # TODO: do this now? or when message is delivered?
        p.acceptInboundMessage(message)

        message.log("will accept rfc822 message: %s" % why)
        message.allowRFC822 = why
        message.allowNewSenderCreation = True

        # other pieces of the message are handled by other processors
        return self.done(message)

    def checkFailed(self, f, message):
        f.trap(NotPermitted)
        why = f.value
        return self.reject(message, why)

    def reject(self, message, why):
        debug = message.agent.debugProcessing
        if debug:
            print "  rejected", why
        message.log("rejected rfc822 message: %s" % why)
        message.allowRFC822 = False
        message.agent.rejectRFC822(message, why)
        return self.done(message)


class CreateSenderEntry(InboundProcessor):
    dependencies = ["checkedNewSenderCreation"]
    provides = ["senderEntry"]

    def done(self, message):
        message.provide("senderEntry")
        return

    def checkMessage(self, message):
        """Make sure there is an address book entry for the sender."""
        InboundProcessor.checkMessage(self, message)

        debug = message.debug

        if message.senderEntry:
            if debug:
                print "CreateSenderEntry: already exists"
            return self.done(message) # already there
        if not message.allowNewSenderCreation:
            if debug:
                print "CreateSenderEntry: not allowed"
            return self.done(message)
        # which addressbook are we using?
        book = message.agent.getAddressBook(message.recipientEntry.ID)
        if not book:
            return self.done(message) # nowhere to put it

        signed = message.getNode("senderIDRecord")
        if not signed:
            # for now, be lazy: if they didn't provide us with their record
            # in their initial message, don't create an entry. TODO?:
            # perhaps do a lookup based upon the ID number or addresses in
            # the record, but then verify that it matches the SenderID in
            # this message.
            if debug:
                print "CreateSenderEntry: they didn't provide an idRecord"
            return self.done(message)

        # must add it to the address book
        entry = addressbook.AddressBookEntry()
        # createIDRecord will import the public key
        d = entry.createIDRecord(signed, message.agent.crypto, message)
        d.addCallback(self._recordUpdated, book, message)
        return d

    def _recordUpdated(self, entry, book, message):
        assert not book.lookupID(entry.idRecord.ID)
        message.senderEntry = entry
        book.addEntry(entry, message)
        message.log("installed AddressBook Entry '%s'" % entry.getName())
        return self.done(message)


class UpdateMyIDRecord(InboundProcessor):
    dependencies = ["senderEntry"]
    provides = ["updatedIDRecord"]

    def done(self, message):
        message.provide("updatedIDRecord")
        return

    def checkMessage(self, message):
        """Update my copy of their ID Record. This is the catcher for the
        update inserted by the UpdateTheirIDRecord outbound processor."""
        InboundProcessor.checkMessage(self, message)

        if not message.senderEntry:
            return self.done(message)
        signed = message.getNode("senderIDRecord")
        if not signed:
            return self.done(message)
        message.log("updating my ID Record")
        d = message.senderEntry.updateIDRecord(signed, message)
        d.addCallback(self._didUpdate, message)
        return d

    def _didUpdate(self, entry, message):
        return self.done(message)

class UpdateOurPermission(InboundProcessor):
    dependencies = ["senderEntry", "updatedIDRecord"]
    provides = ["updatedOurPermission"]

    def done(self, message):
        message.provide("updatedOurPermission")
        return

    def checkMessage(self, message):
        """If the inbound message has a PermissionSlip, verify the signature
        and update the OurPermission field of their address book entry. If
        requested, acknowledge receipt of the PermissionSlip. This is the
        catcher for the UpdateTheirPermission outbound processor.
        """
        InboundProcessor.checkMessage(self, message)

        slip = message.getNode("yourPermission") # TODO: part of rfc822 node?
        if not slip:
            return self.done(message)
        message.log("updating my PermissionSlip")
        d = defer.maybeDeferred(message.senderEntry.updateOurPermission,
                                slip)
        d.addCallback(self._didUpdate, message)
        # TODO: send ack, if requested
        return d

    def _didUpdate(self, result, message):
        return self.done(message)


class AcceptRFC822(InboundProcessor):
    dependencies = ["checkedRFC822", "senderEntry", "updatedIDRecord",
                    "updatedOurPermission"]

    def checkMessage(self, message):
        """Finally accept an incoming rfc822 message."""
        InboundProcessor.checkMessage(self, message)
        a = message.agent
        debug = a.debugProcessing

        if not message.getNode("rfc822"):
            return
        if not message.allowRFC822:
            return

        # tell the agent to deliver the body to the user
        message.log("accepting rfc822 message")
        message.agent.acceptRFC822(message, message.allowRFC822)
        return True

class GrantInboundPermission(InboundProcessor):
    dependencies = ["checkedRFC822", "senderEntry", "updatedIDRecord"]
    permissionGrantHoldoffTime = 60 # TODO: should be more like an hour

    def checkMessage(self, message):
        """If the message was accepted, and not deemed undesireable, grant
        the sender specific permission to send more."""
        InboundProcessor.checkMessage(self, message)

        if not message.senderEntry:
            return
        if not message.allowRFC822:
            return
        entry = message.senderEntry
        if entry.theirPermission:
            return # already have permission
        if entry.doNotGrantReturnPermission:
            message.log("not granting InboundPermission: we don't like them")
            return True
        if not message.recipientEntry.grantInboundPermission:
            message.log("not granting InboundPermission: we don't do that")
            return True

        # TODO: instead of doing this right away, schedule a timer to do it
        # an hour after the message is actually read. Insert a button in the
        # message which will cancel this timer and put a notation in the
        # address book entry that this sender is not to be granted further
        # permission. We'll probably also want to put an expiration timer on
        # the address book entry itself, as the message probably came from a
        # one-off identity

        message.log("will grant them inbound permission in %d seconds" % \
                    self.permissionGrantHoldoffTime)
        why = "accepted mail from new sender, " \
              "granting them permission to send more messages "
        agent = message.agent
        d = agent.scheduleEvent(self.permissionGrantHoldoffTime, why,
                                message.recipientEntry.ID)
        d.addCallback(agent.grantInboundPermission,
                      message.senderEntry.idRecord.ID,
                      message)
        return True # we do not wait for the permission to be added

class GrantThirdPartyPermission(InboundProcessor):
    def checkMessage(self, message):
        """If the message requests that we grant permission to a third
        party, and this capability is enabled in their address book entry,
        send a permission slip to that party."""
        InboundProcessor.checkMessage(self, message)

        # TODO
        return True
        message.log("granted third-party permission")


allInbound = [CheckRFC822Permission,
              CreateSenderEntry,
              UpdateMyIDRecord,
              UpdateOurPermission,
              AcceptRFC822,
              GrantInboundPermission,
              #GrantThirdPartyPermission,
              ]

# outbound:
#  make sure there is a MessageID header node
#  require that the message has a recipientRecord
#  if the message has a recipientAddress but not a recipientRecord, do an
#   address lookup
#  make sure there is a Transport, get one if not
#  if their copy of the ID Record is out of date, include a copy
#  make sure there is a senderID node (or use the crypto envelope for that?)
#  if their permission record is out of date, include a copy
#  rfc822 messages need permission
#  return permission: if we're sending an rfc822 node to them, and they don't
#   have return permission, and the address book entry doesn't specifically
#   say we should deny them permission, then add Permission to their entry
#  3rd-party permission: if the message references other addresses to whom we
#   have permission, ask the user and then request 3rd-party permission for
#   them (keep track of the answer in the address book entry)

# inbound:
#  if they sent us an rfc822 message that was accepted, make sure there is
#   an address book entry for them
#  if they included an ID Record, make sure our copy is up to date (maybe ack)
#  if they included a Permission Slip, update our copy (maybe ack)
#  if there is an rfc822 node, verify the permission, and deliver to the user
#   (maybe ack)
#  if their message was not deemed undesireable, and they don't already have
#   permission, grant it (probably with a configurable timer)
#  if they requested 3rd-party permission, maybe grant it

# other inbound
# for address servers:
#   respond to addressLookup nodes
# for transport servers:
#   respond to transportLookup nodes
